/* $Id: sogen.c,v 1.1.1.1 2001/06/10 12:01:17 masu Exp $
/*--------------------------------------------------------------------
/* Name
/*		sogen.c
/*
/* Description
/*		structre operator code generator
/*
/* Author(s)
/*		masu / Kiyoshi Masumoto <massun.masumoto@nifty.ne.jp>
/*
/*********************************************************************
/**/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "cmdline.h"

/* definition */
#define		STR_SIZE			256			/* string size */
#define		VEC_SIZE			128			/* vector initial size in generated source */

/* static val. */
static char s_sz_include[STR_SIZE];			/* include file name */
static char s_sz_struct[STR_SIZE];			/* structure name */
static char s_sz_struct_low[STR_SIZE];		/* lower case of struct name */
static char s_sz_struct_up[STR_SIZE];		/* upper case of struct name */
static char s_sz_key[STR_SIZE];				/* primary key name */
static int  s_is_key_val = 1;				/* '1' means that key type is value. '0' means that key type is string */
static int  s_is_key_assigned = 0;			/* '1': key exists. 0: not defined key */
static char s_sz_error_msg[STR_SIZE];		/* error message string */

/* prototype decl. */
/* static function */
static int gen_include( const char * );
static int gen_c( const char * );

static void gen_prototype( FILE *fp, const char *sz_prefix );
static void gen_struct_copy( FILE * );
static void gen_vector_extend( FILE * );
static void gen_compare( FILE * );
/* .. container operation */
static void gen_push_back( FILE * );
static void gen_pop_back( FILE * );
static void gen_touch_pop_back( FILE * );
static void gen_clear( FILE * );
static void gen_touch_clear( FILE * );
static void gen_size( FILE * );
static void gen_at( FILE * );
/* .. find operation */
static void gen_find( FILE * );
static void gen_findfirst( FILE * );
static void gen_findnext( FILE * );
static void gen_bsearch( FILE * );
static void gen_bsearch_by( FILE * );
/* .. sort operation */
static void gen_sort( FILE * );
static void gen_sort_by( FILE * );
/* .. transform operation */
static void gen_transform( FILE * );
static void gen_transform_if( FILE * );

static int  is_strlen_overflow( const char * );


/*********************************************************************
/* Descriptoion
/*		make lower string
/* In
/*		p_src : const char * ( input string )
/* Out
/*		p_dst : char * ( to lowwer string )
/* Return
/*		none
/*********************************************************************
/**/
static void
lower_string( char *p_dst, const char *p_src )
{
	unsigned int ui_size = strlen( p_src );
	unsigned int i;
	for ( i=0; i < ui_size ; i++ ) {
		p_dst[i] = ( char )( tolower( p_src[i] ) );
	}
	p_dst[i] = '\0';
}


/*********************************************************************
/* Descriptoion
/*		make upper string
/* In
/*		p_src : const char * ( input string )
/* Out
/*		p_dst : char * ( to upper string )
/* Return
/*		none
/*********************************************************************
/**/
static void
upper_string( char *p_dst, const char *p_src )
{
	unsigned int ui_size = strlen( p_src );
	unsigned int i;
	for ( i=0; i < ui_size ; i++ ) {
		p_dst[i] = ( char )( toupper( p_src[i] ) );
	}
	p_dst[i] = '\0';
}


/*********************************************************************
/* Description
/*		structure operator code generator
/* In
/*		sz_include  : const char *  ( include filename )
/*		sz_struct   : const char *  ( structure name )
/*		sz_key      : const char *  ( key string in structre )
/*		is_key_val  : int ( !=0:value, ==0:string )
/* Out
/*		none
/* Return
/*		== 0 : success
/*		!= 0 : error
/*********************************************************************
/**/
int
sogen_code_generate(const char * sz_include,
					const char * sz_struct,
					const char * sz_key,
					int is_key_val )
{
	/* copy to static member */
	if (   is_strlen_overflow( sz_include )
		|| is_strlen_overflow( sz_struct )
		|| is_strlen_overflow( sz_key ) ) {
		/* error */
		sprintf( s_sz_error_msg, "string is too long.\n" );
		return 1;
	}
	strcpy( s_sz_include, sz_include );
	strcpy( s_sz_struct, sz_struct );
	lower_string( s_sz_struct_low, sz_struct );
	upper_string( s_sz_struct_up, sz_struct );
	strcpy( s_sz_key, sz_key );
	if ( *s_sz_key != '\0' ) s_is_key_assigned = 1;
	s_is_key_val = is_key_val;

	/* include file generate
	/*	<struct>_mgr.h
	/**/
	{
		char sz_filename[FILENAME_MAX];
		sprintf( sz_filename, "%s%s.h", s_sz_struct_low, cmdline_get_string_body_suffix() );
		if ( gen_include( sz_filename ) ) {
			/* error */
			return 1;
		}
	}

	/* implimentation file generate 
	/*	<struct>_mgr.c
	/**/
	{
		char sz_filename[FILENAME_MAX];
		sprintf( sz_filename, "%s%s.c", s_sz_struct_low, cmdline_get_string_body_suffix() );
		if ( gen_c( sz_filename ) ) {
			/* error */
			return 1;
		}
	}

	return 0;
}


/*********************************************************************
/* Description
/*		generate inlude file
/* In
/*		sz_include : const char *
/* Out
/*		none
/* Return
/*		== 0 : success
/*		!= 0 : error
/*********************************************************************
/**/
static int
gen_include( const char *sz_include )
{
	FILE *fp;
	char body_suffix_up[STR_SIZE];	/* body suffix. e.g. "_mgr" */

	fp = fopen( sz_include, "w" );
	if ( fp == NULL ) {
		sprintf( s_sz_error_msg, "file open error.(%s)\n", sz_include );
		return 1;
	}

	upper_string( body_suffix_up, cmdline_get_string_body_suffix() );

	fprintf( fp, "#ifndef %s%s_H\n", s_sz_struct_up, body_suffix_up );
	fprintf( fp, "#define %s%s_H 1\n", s_sz_struct_up, body_suffix_up );
	fprintf( fp, "#include \"%s\"\n", s_sz_include );
	fprintf( fp, "\n" );

	gen_prototype( fp, "extern " );

	fprintf( fp, "\n" );
	fprintf( fp, "#endif  /* !%s%s_H */\n", s_sz_struct_up, body_suffix_up );

	fclose( fp );

	return 0;
}


/*********************************************************************
/* Description
/*		generate c file
/* In
/*		sz_file : const char *
/* Out
/*		none
/* Return
/*		== 0 : success
/*		!= 0 : error
/*********************************************************************
/**/
static int
gen_c( const char *sz_file )
{
	FILE *fp;

	fp = fopen( sz_file, "w" );
	if ( fp == NULL ) {
		sprintf( s_sz_error_msg, "file open error.(%s)\n", sz_file );
		return 1;
	}

	/* include */
	fprintf( fp, "#include <stdio.h>\n" );
	fprintf( fp, "#include <stdlib.h>\n" );
	fprintf( fp, "#include <string.h>\n" );
	fprintf( fp, "#include \"%s\"\n", s_sz_include );
	fprintf( fp, "\n" );

	/* definition */
	fprintf( fp, "#define       SIZE_VECTOR        %d\n", VEC_SIZE );
	fprintf( fp, "\n" );

	/* static val. */
	fprintf( fp, "/* vector management */\n" );
	fprintf( fp, "static %s *p_vec_top = NULL;\n", s_sz_struct );
	fprintf( fp, "static size_t n_vec_size = 0;\n" );
	fprintf( fp, "static size_t n_vec_max_size= 0;\n" );
	fprintf( fp, "/* hold found position */\n" );
	fprintf( fp, "static size_t n_cur_index = 0;\n" );
	fprintf( fp, "\n" );

	/* protorype decl. */
	gen_prototype( fp, "" );
	fprintf( fp, "\n" );

	/* static function */
	gen_struct_copy( fp );
	gen_vector_extend( fp );
	gen_compare( fp );
	fprintf( fp, "\n" );

	/* body */
	gen_push_back( fp );
	gen_pop_back( fp );
	gen_touch_pop_back( fp );
	gen_clear( fp );
	gen_touch_clear( fp );
	gen_size( fp );
	gen_at( fp );
	gen_find( fp );
	gen_findfirst( fp );
	gen_findnext( fp );
	gen_bsearch( fp );
	gen_bsearch_by( fp );
	gen_sort( fp );
	gen_sort_by( fp );
	gen_transform( fp );
	gen_transform_if( fp );

	fprintf( fp, "\n" );

	fclose( fp );
	return 0;
}


/*********************************************************************
/* Description
/*		outut protorype decl. to file .
/* In
/*		fp : FILE *
/*		sz_prefix : const char * ( e.g. "extern "
/* Out
/*		none
/* Return
/*		none
/*********************************************************************
/**/
static void
gen_prototype( FILE *fp, const char *sz_prefix )
{
	fprintf( fp, "/* container operation */\n" );
	fprintf( fp, "%sint %s_push_back( const %s * );\n", sz_prefix, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%svoid %s_pop_back( void );\n", sz_prefix, s_sz_struct_low );
	fprintf( fp, "%svoid %s_touch_pop_back( void ( *fn )( %s * ) );\n", sz_prefix, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%svoid %s_clear( void );\n" , sz_prefix, s_sz_struct_low );
	fprintf( fp, "%svoid %s_touch_clear( void ( *fn )( %s * ) );\n" , sz_prefix, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%ssize_t %s_size( void );\n", sz_prefix, s_sz_struct_low );
	fprintf( fp, "%sconst %s * %s_at( size_t );\n", sz_prefix, s_sz_struct, s_sz_struct_low );
	fprintf( fp, "\n" );

	fprintf( fp, "/* find operation */\n" );
	fprintf( fp, "%sint %s_find( %s ** );\n", sz_prefix, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%sint %s_findfirst( int ( *fn_flt )( %s * ), %s ** );\n", sz_prefix, s_sz_struct_low, s_sz_struct, s_sz_struct );
	fprintf( fp, "%sint %s_findnext( int ( *fn_flt )( %s * ), %s ** );\n", sz_prefix, s_sz_struct_low, s_sz_struct, s_sz_struct );
	if ( s_is_key_assigned )
		fprintf( fp, "%s%s * %s_bsearch( %s * );\n", sz_prefix, s_sz_struct, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%s%s * %s_bsearch_by( %s *, int ( *fn_cmp )( const void * ,const void * ) );\n", sz_prefix, s_sz_struct, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "\n" );

	fprintf( fp, "/* sort operation */\n" );
	if ( s_is_key_assigned )
		fprintf( fp, "%svoid %s_sort( void );\n", sz_prefix, s_sz_struct_low );
	fprintf( fp, "%svoid %s_sort_by( int ( *fn_cmp )( const void * ,const void * ) );\n", sz_prefix, s_sz_struct_low );
	fprintf( fp, "\n" );

	fprintf( fp, "/* transform operation */\n" );
	fprintf( fp, "%svoid %s_transform( void ( *fn )( %s * ) );\n", sz_prefix, s_sz_struct_low, s_sz_struct );
	fprintf( fp, "%svoid %s_transform_if( int ( *fn_flt )( %s * ), void ( *fn )( %s * ) );\n", sz_prefix, s_sz_struct_low, s_sz_struct, s_sz_struct );
}


/*********************************************************************
/* Description
/*		generate function <name>_push_back()
/*		 that push back element to container.
/* Out
/*		fp : FILE *fp
/* Return 
/*		none 
/*********************************************************************
/**/
static void
gen_push_back( FILE *fp )
{
	fprintf( fp, "int\n" );
	fprintf( fp, "%s_push_back( const %s *pst_%s )\n", s_sz_struct_low, s_sz_struct, s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tif ( vector_extend() ) return -1;\n" );
	fprintf( fp, "\tstruct_copy( &p_vec_top[ n_vec_size ], pst_%s );\n", s_sz_struct_low );
	fprintf( fp, "\tn_vec_size++;\n" );
	fprintf( fp, "\treturn 0;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_pop_back()
/*		that pop back element from container.
/* Out
/*		fp : FILE *
/* Return 
/*		none 
/*********************************************************************
/**/
static void
gen_pop_back( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_pop_back( void )\n", s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tif ( n_vec_size > 0 ) {\n" );
	fprintf( fp, "\t\tn_vec_size--;\n" );
	fprintf( fp, "\t}\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_touch_pop_back()
/*		that touch and pop back element from container.
/* Out
/*		fp : FILE *
/* Return 
/*		none 
/*********************************************************************
/**/
static void
gen_touch_pop_back( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_touch_pop_back( void ( *fn )( %s * ) )\n", s_sz_struct_low, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tif ( n_vec_size > 0 ) {\n" );
	fprintf( fp, "\t\tif ( fn ) { ( *fn )( &p_vec_top[n_vec_size-1] ); }\n" );
	fprintf( fp, "\t\tn_vec_size--;\n" );
	fprintf( fp, "\t}\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_clear
/*		that clear container.
/* Out
/*		fp : FILE *
/* Return
/*		none
/*********************************************************************
/**/
static void
gen_clear( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_clear( void )\n", s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "%s_touch_clear( NULL );\n", s_sz_struct_low );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_touch_clear
/*		that touch and clear container.
/* Out
/*		fp : FILE *
/* Return
/*		none
/*********************************************************************
/**/
static void
gen_touch_clear( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_touch_clear( void ( *fn )( %s * ) )\n", s_sz_struct_low, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tint i_cnt = %s_size();\n", s_sz_struct_low );
	fprintf( fp, "\tint i;\n" );
	fprintf( fp, "\tfor ( i=0; i < i_cnt; i++ )\n" );
	fprintf( fp, "\t\t%s_touch_pop_back( fn );\n", s_sz_struct_low );
	fprintf( fp, "\n" );
	fprintf( fp, "\tfree( p_vec_top );\n" );
	fprintf( fp, "\tp_vec_top = NULL;\n" );
	fprintf( fp, "\tn_vec_max_size = 0;\n" );

	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_size()
/*		that get container size.
/* Out
/*		fp : FILE *
/* Return
/*		none
/*********************************************************************
/**/
static void
gen_size( FILE *fp )
{
	fprintf( fp, "size_t\n" );
	fprintf( fp, "%s_size( void )\n", s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "\treturn n_vec_size;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_at()
/*		that get element assigned by index.
/* Out
/*		fp : FILE *
/* Return
/*		none
/*********************************************************************
/**/
static void
gen_at( FILE *fp )
{
	fprintf( fp, "const %s *\n", s_sz_struct );
	fprintf( fp, "%s_at( size_t i )\n", s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tif ( i >= %s_size() ) return NULL;\n", s_sz_struct_low );
	fprintf( fp, "\treturn &p_vec_top[i];\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_find()
/*		that find element
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_find( FILE *fp )
{
	fprintf( fp, "int\n" );
	fprintf( fp, "%s_find( %s **pp )\n", s_sz_struct_low, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\treturn %s_findfirst( NULL, pp );\n", s_sz_struct_low );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_findfirst()
/*		that find element filted conditon .
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_findfirst( FILE *fp )
{
	fprintf( fp, "int\n" );
	fprintf( fp, "%s_findfirst( int ( *fn_flt )( %s * ), %s **pp )", s_sz_struct_low, s_sz_struct, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tsize_t i;\n" );
	fprintf( fp, "\t%s *p_%s = p_vec_top;\n", s_sz_struct, s_sz_struct_low );
	fprintf( fp, "\tfor ( i = 0; i < n_vec_size; i++ ) {\n" );
	fprintf( fp, "\t\tif (   fn_flt == NULL\n" );
	fprintf( fp, "\t\t\t|| ( *fn_flt )( &p_%s[i] ) ) {\n", s_sz_struct_low );
	fprintf( fp, "\t\t\t*pp = &p_%s[i];\n", s_sz_struct_low );
	fprintf( fp, "\t\t\tn_cur_index = i;\n" );
	fprintf( fp, "\t\t\treturn 1;\n" );
	fprintf( fp, "\t\t}\n" );
	fprintf( fp, "\t}\n" );
	fprintf( fp, "\treturn 0;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_findnext()
/*		that find element filted conditon .
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_findnext( FILE *fp )
{
	fprintf( fp, "int\n" );
	fprintf( fp, "%s_findnext( int ( *fn_flt )( %s * ), %s **pp )", s_sz_struct_low, s_sz_struct, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tsize_t i;\n" );
	fprintf( fp, "\t%s *p_%s = p_vec_top;\n", s_sz_struct, s_sz_struct_low );
	fprintf( fp, "\tfor ( i = n_cur_index+1; i < n_vec_size; i++ ) {\n" );
	fprintf( fp, "\t\tif (   fn_flt == NULL\n" );
	fprintf( fp, "\t\t\t|| ( *fn_flt )( &p_%s[i] ) ) {\n", s_sz_struct_low );
	fprintf( fp, "\t\t\t*pp = &p_%s[i];\n", s_sz_struct_low );
	fprintf( fp, "\t\t\tn_cur_index = i;\n" );
	fprintf( fp, "\t\t\treturn 1;\n" );
	fprintf( fp, "\t\t}\n" );
	fprintf( fp, "\t}\n" );
	fprintf( fp, "\treturn 0;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_bsearch()
/*		that search element.
/*		can't be made this function, if key is assigned or, key type is string.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_bsearch( FILE *fp )
{
	if ( s_is_key_assigned ) {
		fprintf( fp, "%s *\n", s_sz_struct );
		fprintf( fp, "%s_bsearch( %s *p_key )\n", s_sz_struct_low, s_sz_struct );
		fprintf( fp, "{\n" );
		fprintf( fp, "\treturn ( %s *)bsearch( p_key, p_vec_top, n_vec_size, sizeof( %s ), compare );\n", s_sz_struct, s_sz_struct );
		fprintf( fp, "}\n" );
		fprintf( fp, "\n" );
	}
}


/*********************************************************************
/* Description
/*		generate function <name>_bsearch_by()
/*		that search element filted conditon .
/*		can't be made this function, if key is assigned or, key type is string.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_bsearch_by( FILE *fp )
{
	fprintf( fp, "%s *\n", s_sz_struct );
	fprintf( fp, "%s_bsearch_by( %s *p_key, int ( *fn_cmp )( const void * , const void * ) )\n", s_sz_struct_low, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\treturn ( %s *)bsearch( p_key, p_vec_top, n_vec_size, sizeof( %s ), fn_cmp );\n", s_sz_struct, s_sz_struct );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_sort()
/*		that sort elements in container.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_sort( FILE *fp )
{
	if ( s_is_key_assigned ) {
		fprintf( fp, "void\n" );
		fprintf( fp, "%s_sort( void )\n", s_sz_struct_low );
		fprintf( fp, "{\n" );
		fprintf( fp, "\tqsort( p_vec_top, n_vec_size, sizeof( %s ), compare );\n", s_sz_struct );
		fprintf( fp, "}\n" );
		fprintf( fp, "\n" );
	}
}


/*********************************************************************
/* Description
/*		generate function <name>_sort_by()
/*		that sort elements in container by condition.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_sort_by( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_sort_by( int ( *fn_cmp )( const void * , const void * ) )\n", s_sz_struct_low );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tqsort( p_vec_top, n_vec_size, sizeof( %s ), fn_cmp );\n", s_sz_struct );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_transform()
/*		that transform elements.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_transform( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_transform( void ( *fn )( %s * ) )\n", s_sz_struct_low, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\t%s_transform_if( NULL, fn );\n", s_sz_struct_low );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function <name>_transform()
/*		that transform elements by condition
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_transform_if( FILE *fp )
{
	fprintf( fp, "void\n" );
	fprintf( fp, "%s_transform_if( int ( *fn_flt )( %s * ), void ( *fn )( %s * ) )\n", s_sz_struct_low, s_sz_struct, s_sz_struct );
	fprintf( fp, "{\n" );

	fprintf( fp, "\tsize_t i;\n" );
	fprintf( fp, "\tfor ( i = 0; i < n_vec_size; i++ ) {\n" );
	fprintf( fp, "\t\tif (   fn_flt == NULL\n" );

	fprintf( fp, "\t\t\t|| ( *fn_flt )( &p_vec_top[i] ) ) {\n" );
	fprintf( fp, "\t\t\t( *fn )( &p_vec_top[i] );\n" );
	fprintf( fp, "\t\t}\n" );
	fprintf( fp, "\t}\n" );

	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function static <name>_struct_copy()
/*		that is default struct copy operaton.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_struct_copy( FILE *fp )
{
	fprintf( fp, "static void\n" );
	fprintf( fp, "struct_copy( %s *p_dst, const %s *p_src )\n", s_sz_struct, s_sz_struct );
	fprintf( fp, "{\n" );
	fprintf( fp, "\t*p_dst = *p_src;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function static <name>_vector_extend()
/*		that extend vector.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_vector_extend( FILE *fp )
{
	fprintf( fp, "static int\n" );
	fprintf( fp, "vector_extend( void )\n" );
	fprintf( fp, "{\n" );
	fprintf( fp, "\tif ( n_vec_size + 1 >= n_vec_max_size ) {\n" );
	fprintf( fp, "\t\tif ( n_vec_max_size == 0 ) {\n" );
	fprintf( fp, "\t\t\tn_vec_max_size = SIZE_VECTOR;\n" );
	fprintf( fp, "\t\t\tp_vec_top = ( %s * )calloc( n_vec_max_size, sizeof( %s ) );\n", s_sz_struct, s_sz_struct );
	fprintf( fp, "\t\t\tif ( p_vec_top == NULL ) {\n" );
	fprintf( fp, "\t\t\t\tfprintf( stderr, \"memory allocation error.( vector_extend() )\\n\" );\n" );
	fprintf( fp, "\t\t\t\treturn -1;\n" );
	fprintf( fp, "\t\t\t}\n" );
	fprintf( fp, "\t\t} else {\n" );
	fprintf( fp, "\t\t\tn_vec_max_size *= 2;\n" );
	fprintf( fp, "\t\t\tif ( !realloc( p_vec_top, sizeof( %s ) * n_vec_max_size ) ) {\n", s_sz_struct );
	fprintf( fp, "\t\t\t\tfprintf( stderr, \"memory allocation error.( vector_extend() )\\n\" );\n" );
	fprintf( fp, "\t\t\t\treturn -1;\n" );
	fprintf( fp, "\t\t\t}\n" );
	fprintf( fp, "\t\t}\n" );
	fprintf( fp, "\t}\n" );
	fprintf( fp, "\treturn 0;\n" );
	fprintf( fp, "}\n" );
	fprintf( fp, "\n" );
}


/*********************************************************************
/* Description
/*		generate function static <name>_cmpare()
/*		that is default compare function.
/* Out
/*		fp : FILE *
/* Return 
/*		none
/*********************************************************************
/**/
static void
gen_compare( FILE *fp )
{
	if ( s_is_key_assigned ) {
		fprintf( fp, "/* default compare function */\n" );
		fprintf( fp, "static int\n" );
		fprintf( fp, "compare( const void *p1, const void *p2 )\n" );
		fprintf( fp, "{\n" );
		if ( s_is_key_val ) {
			/* case of value */
			fprintf( fp, "\tif ( ((%s *)p1)->%s > ((%s *)p2)->%s )\n", s_sz_struct, s_sz_key, s_sz_struct, s_sz_key );
			fprintf( fp, "\t\treturn 1;\n" );
			fprintf( fp, "\tif ( ((%s *)p1)->%s < ((%s *)p2)->%s )\n", s_sz_struct, s_sz_key, s_sz_struct, s_sz_key );
			fprintf( fp, "\t\treturn -1;\n" );
			fprintf( fp, "\treturn 0;\n" );
		} else {
			/* case of string */
			fprintf( fp, "\treturn strcmp( ((%s *)p1)->%s, ((%s *)p2)->%s );\n", s_sz_struct, s_sz_key, s_sz_struct, s_sz_key );
		}
		fprintf( fp, "}\n" );
		fprintf( fp, "\n" );
	}
}


/*********************************************************************
/* Description
/*		check string size
/* In
/*		s : checking string
/* Return 
/*		== 0 : OK
/*		!= 1 : overflow ( too long )
/*********************************************************************
/**/
static int
is_strlen_overflow( const char *s )
{
	if ( strlen( s ) >= STR_SIZE ) {
		return 1;
	}
	return 0;
}
